/*
  ICFS: Interactively Controlled File System
  Copyright (C) 2024-2025  Fedir Kovalov

  This program can be distributed under the terms of the GNU GPLv2.
  See the file LICENSE.
*/

#include <stddef.h>
#include <sys/types.h>
#include <time.h>
#define _GNU_SOURCE
#include "cc.h"
#include "ui-socket.h"
#include <errno.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>

/**
 * Function to get the process creation time (in jiffies) from the proc
 * filesystem
 *
 * @param pid: The process ID of the process to get the creation time of
 * @return: The process creation time in jiffies, or 0 on error
 * @note: although nothing in the documentation says that the creation time is
 * never really equal to 0, it exceptionally unlikely.
 */
unsigned long long get_process_creation_time(pid_t pid) {
  char path[32];
  FILE *fp;
  unsigned long long creation_time = 0;

  // Construct the path to the process's status file
  snprintf(path, sizeof(path), "/proc/%d/stat", pid);

  // Open the status file
  fp = fopen(path, "r");
  if (fp == NULL) {
    perror("fopen");
    return 0;
  }

  // Read the creation time (the 22nd field in the stat file)
  for (int i = 1; i < 22; i++) {
    if (fscanf(fp, "%*s") != 1) {
      fprintf(stderr, "Error reading process stat file\n");
      fclose(fp);
      return 0;
    }
  }
  if (fscanf(fp, "%llu", &creation_time) != 1) {
    fprintf(stderr, "Error reading creation time\n");
    fclose(fp);
    return 0;
  }

  // Close the file
  fclose(fp);

  return creation_time;
}

struct temp_process_permissions {
  // yes, this is a correct type for start time in jiffies (see
  // proc_pid_stat(5))
  unsigned long long creation_time;
  vec(char *) allowed_files;
};

map(pid_t, struct temp_process_permissions) temp_permissions_table;
pthread_mutex_t temp_permissions_table_lock;

int init_ui_socket() {
  char line[256];
  FILE *fp;

  // Test if Zenity is installed (get version)
  fp = popen("zenity --version", "r");
  if (fp == NULL) {
    perror("Pipe returned a error");
    return 1;
  }
  pthread_mutex_init(&temp_permissions_table_lock, PTHREAD_MUTEX_DEFAULT);
  init(&temp_permissions_table);

  while (fgets(line, sizeof(line), fp))
    printf("%s", line);
  pclose(fp);
  return 0;
}

void destroy_ui_socket() {
  // free the memory allocated for the table
  for_each(&temp_permissions_table, entry) {
    for_each(&entry->allowed_files, allowed_file) { free(*allowed_file); }
    cleanup(&entry->allowed_files);
  }
  cleanup(&temp_permissions_table);
  pthread_mutex_destroy(&temp_permissions_table_lock);
}

/**
 * Asks the user if the process should be allowed to access the file using the
 * GUI
 *
 * @param filename: The file that the process is trying to access
 * @pram pi: The process information
 * @return: 0 if access is denied, 1 if access is allowed, 2 if access is allwed
 * for the runtime of the process
 */
int ask_access(const char *filename, struct process_info pi) {

  FILE *fp;
  size_t command_len =
      139 + sizeof(pid_t) * 8 + strlen(pi.name) + strlen(filename);
  char *command = (char *)malloc(command_len);
  snprintf(command, command_len,
           "zenity --question --extra-button \"Allow this time\" --title "
           "\"Allow Access?\" --text \"Allow process "
           "<tt>%s</tt> with PID <tt>%d</tt> to access <tt>%s</tt>\"",
           pi.name, pi.PID, filename);

  // Zenity Question Message Popup
  fp = popen(command, "r");
  free(command);

  if (fp == NULL) {
    perror("Pipe returned a error");
    return -1;
  }

  // if the user clicks the "Allow this time" button, `zenity` will only
  // write it to `stdout`, but the exit code will still be `1`. So, we need
  // to manually check the output.
  char buffer[1024];
  while (fgets(buffer, sizeof(buffer), fp)) {
    printf("%s", buffer);
    if (strcmp(buffer, "Allow this time\n") == 0) {
      pclose(fp);
      return 2;
    }
  }

  int zenity_exit_code = WEXITSTATUS(pclose(fp));
  // zenity returns 1 on "No" >:(
  if (zenity_exit_code == 0) {
    return 1;
  }

  return 0;
}

/**
 * Checks if the process has a temporary access to the file.
 *
 * @param filename: The file that the process is trying to access
 * @pram pi: The process information
 * @return: 0 if access is denied, 1 if access is allowed
 */
int check_temp_access(const char *filename, struct process_info pi) {
  // TODO: more efficient locking
  pthread_mutex_lock(&temp_permissions_table_lock);
  struct temp_process_permissions *permission_entry =
      get(&temp_permissions_table, pi.PID);
  if (permission_entry != NULL) {
    unsigned long long process_creation_time =
        get_process_creation_time(pi.PID);
    if (process_creation_time == 0) {
      perror("Could not retrieve process creation time");
      pthread_mutex_unlock(&temp_permissions_table_lock);
      return 0;
    }

    if (process_creation_time == permission_entry->creation_time) {
      // the process is the same as the one that was granted temporary access
      // to the file
      for_each(&permission_entry->allowed_files, allowed_file) {
        if (strncmp(*allowed_file, filename, strlen(filename)) == 0) {
          pthread_mutex_unlock(&temp_permissions_table_lock);
          return 1;
        }
      }
    }
  }
  pthread_mutex_unlock(&temp_permissions_table_lock);
  return 0;
}

/**
 * Checks if the process has a permanent access to the file.
 *
 * @param filename: The file that the process is trying to access
 * @pram pi: The process information
 * @return: 0 if access is denied, 1 if access is allowed
 */
int check_perm_access(const char *filename, struct process_info pi) {
  perror("Not implemented");
  return 0;
}

int give_temp_access(const char *filename, struct process_info pi) {
  pthread_mutex_lock(&temp_permissions_table_lock);
  struct temp_process_permissions *permission_entry =
      get(&temp_permissions_table, pi.PID);

  if (permission_entry != NULL) {

    unsigned long long process_creation_time =
        get_process_creation_time(pi.PID);
    if (process_creation_time == 0) {
      perror("Could not retrieve process creation time");
      pthread_mutex_unlock(&temp_permissions_table_lock);
      return -1;
    }

    if (process_creation_time == permission_entry->creation_time) {
      // the process is the same as the one that was granted temporary access
      // to the file
      push(&permission_entry->allowed_files, strdup(filename));
      pthread_mutex_unlock(&temp_permissions_table_lock);
      return 0;
    }
    // we have an entry for the process, but the process is different
    // delete the entry and create a new one
    erase(&temp_permissions_table, pi.PID);
    permission_entry = NULL;
  }

  // no entry is present
  // construct the entry
  struct temp_process_permissions new_permission_entry;

  new_permission_entry.creation_time = get_process_creation_time(pi.PID);
  init(&new_permission_entry.allowed_files);
  push(&new_permission_entry.allowed_files, strdup(filename));

  insert(&temp_permissions_table, pi.PID, new_permission_entry);

  pthread_mutex_unlock(&temp_permissions_table_lock);
  return 0;
}

int give_perm_access(const char *filename, struct process_info pi) {
  perror("Not implemented");
  return -1;
}

/**
 * Check access according to:
 * 1. temp permission table
 * 2. permanent permission table
 * 3. user descision
 *
 * @param filename: The file that the process is trying to access
 * @pram pi: The process information
 * @return: 0 if access is denied, 1 if access is allowed
 */
int interactive_access(const char *filename, struct process_info pi) {

  if (check_temp_access(filename, pi) || check_perm_access(filename, pi)) {
    // access was already granted before
    return 1;
  }

  int user_response = ask_access(filename, pi);
  if (user_response == 1) {
    // user said "yes"
    give_perm_access(filename, pi);
    return 1;
  } else if (user_response == 2) {
    // user said "yes, but only this time"
    give_temp_access(filename, pi);
    return 1;
  }

  // otherwise "no"
  return 0;
}
